package de.unikoblenz.fgbks.base.utils.boundary.impl;

import static de.unikoblenz.fgbks.base.utils.boundary.bicreater.BoundaryBiCreaterType.APPEND;
import static de.unikoblenz.fgbks.base.utils.boundary.bicreater.BoundaryBiCreaterType.BETWEEN;
import static de.unikoblenz.fgbks.base.utils.boundary.bicreater.BoundaryBiCreaterType.COMBINE;
import static de.unikoblenz.fgbks.base.utils.boundary.bicreater.BoundaryBiCreaterType.INTERSECTION;
import static de.unikoblenz.fgbks.base.utils.boundary.bicreater.BoundaryBiCreaterType.LOWER_BOUNDS;
import static de.unikoblenz.fgbks.base.utils.boundary.bicreater.BoundaryBiCreaterType.UPPER_BOUNDS;
import static de.unikoblenz.fgbks.base.utils.boundary.checker.BoundaryCheckType.IS_EQUAL;
import static de.unikoblenz.fgbks.base.utils.boundary.checker.BoundaryCheckType.IS_IN_CONTACT;
import static de.unikoblenz.fgbks.base.utils.boundary.checker.BoundaryCheckType.IS_NOT_IN_CONTACT;
import static de.unikoblenz.fgbks.base.utils.boundary.checker.BoundaryCheckType.IS_OVERLAPPING;
import static de.unikoblenz.fgbks.base.utils.boundary.checker.BoundaryCheckType.IS_SUBSUMPTION;
import static de.unikoblenz.fgbks.base.utils.boundary.checker.BoundaryCheckType.SUBSUMES;
import static de.unikoblenz.fgbks.base.utils.boundary.creater.BoundaryCreaterType.LOWER;
import static de.unikoblenz.fgbks.base.utils.boundary.creater.BoundaryCreaterType.UPPER;

import de.unikoblenz.fgbks.base.utils.boundary.AbstractBoundary;
import de.unikoblenz.fgbks.base.utils.boundary.Boundary;
import de.unikoblenz.fgbks.base.utils.boundary.BoundaryParseException;
import de.unikoblenz.fgbks.base.utils.boundary.bicreater.BoundaryBiCreater;
import de.unikoblenz.fgbks.base.utils.boundary.bicreater.BoundaryBiCreaterType;
import de.unikoblenz.fgbks.base.utils.boundary.bicreater.EmptyBiCreater;
import de.unikoblenz.fgbks.base.utils.boundary.checker.BoundaryCheck;
import de.unikoblenz.fgbks.base.utils.boundary.checker.BoundaryCheckType;
import de.unikoblenz.fgbks.base.utils.boundary.checker.CheckStringEqual;
import de.unikoblenz.fgbks.base.utils.boundary.checker.CheckStringInContact;
import de.unikoblenz.fgbks.base.utils.boundary.checker.CheckStringNotInContact;
import de.unikoblenz.fgbks.base.utils.boundary.checker.CheckStringOverlap;
import de.unikoblenz.fgbks.base.utils.boundary.checker.CheckStringSubsumes;
import de.unikoblenz.fgbks.base.utils.boundary.checker.CheckStringSubsumption;
import de.unikoblenz.fgbks.base.utils.boundary.creater.BoundaryCreater;
import de.unikoblenz.fgbks.base.utils.boundary.creater.BoundaryCreaterType;
import de.unikoblenz.fgbks.base.utils.boundary.creater.CreaterLower;
import de.unikoblenz.fgbks.base.utils.boundary.creater.CreaterUpper;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Objects;
import org.apache.commons.lang3.Validate;

public class StringBoundary extends AbstractBoundary<String> {

  private static HashMap<BoundaryCheckType, BoundaryCheck> checkerMap = new HashMap<>();
  private static HashMap<BoundaryCreaterType, BoundaryCreater> createrMap = new HashMap<>();
  private static HashMap<BoundaryBiCreaterType, BoundaryBiCreater> combinerMap = new HashMap<>();

  static {
    checkerMap.put(IS_EQUAL, CheckStringEqual.getInstance());
    checkerMap.put(IS_IN_CONTACT, CheckStringInContact.getInstance());
    checkerMap.put(IS_NOT_IN_CONTACT, CheckStringNotInContact.getInstance());
    checkerMap.put(IS_OVERLAPPING, CheckStringOverlap.getInstance());
    checkerMap.put(IS_SUBSUMPTION, CheckStringSubsumption.getInstance());
    checkerMap.put(SUBSUMES, CheckStringSubsumes.getInstance());
    createrMap.put(UPPER, CreaterUpper.getInstance());
    createrMap.put(LOWER, CreaterLower.getInstance());
    combinerMap.put(INTERSECTION, EmptyBiCreater.getInstance());
    combinerMap.put(LOWER_BOUNDS, EmptyBiCreater.getInstance());
    combinerMap.put(UPPER_BOUNDS, EmptyBiCreater.getInstance());
    combinerMap.put(BETWEEN, EmptyBiCreater.getInstance());
    combinerMap.put(APPEND, EmptyBiCreater.getInstance());
    combinerMap.put(COMBINE, EmptyBiCreater.getInstance());
  }

  private boolean matchNoneOfValues;
  private boolean matchAny;
  private String[] values;
  // sorted list of cached hashes of valuesHashes
  private int[] valuesHashes;

  public StringBoundary(String text) {
    super(text, String.class);
  }

  @Override
  public int compareTo(Boundary o) {
    return o.getText().compareTo(o.getText());
  }

  public boolean matchesNoneOfValues() {
    return matchNoneOfValues;
  }

  public boolean isMatchAny() {
    return matchAny;
  }

  public String[] getValues() {
    return values.clone(); // performance ?
  }

  public int[] getValuesHashes() {
    return valuesHashes.clone(); // performance ?
  }

  public int getAmountOfElements() {
    return values.length;
  }

  @Override
  protected void parse(String text) {
    matchNoneOfValues = false;
    matchAny = false;
    try {
      matchAny = Validate.notNull(text).isEmpty();
      if (matchAny) {
        values = new String[0];
        valuesHashes = new int[0];
      } else {
        matchNoneOfValues = text.toLowerCase().startsWith("not");
        text = text.replace('(', '\0').replace(')', '\0').trim();
        if (matchNoneOfValues) {
          text = text.replaceFirst("not", "").trim();
        }
        values = text.split(",");
        valuesHashes = new int[values.length];
        for (int i = 0; i < values.length; i++) {
          values[i] = values[i].replace('"', '\0').trim();
          valuesHashes[i] = Objects.hashCode(values[i]);
        }
      }
      // Sort hashes for later comparisons
      Arrays.sort(valuesHashes);
    } catch (Exception e) {
      throw new BoundaryParseException("String value is not valid: " + this.getText());
    }
  }

  @Override
  protected void validate() {
    Validate.noNullElements(values);
  }

  @Override
  protected HashMap<BoundaryCheckType, BoundaryCheck> getChecker() {
    return checkerMap;
  }

  @Override
  protected HashMap<BoundaryBiCreaterType, BoundaryBiCreater> getBiCreater() {
    return combinerMap;
  }

  @Override
  protected HashMap<BoundaryCreaterType, BoundaryCreater> getCreater() {
    return createrMap;
  }

  @Override
  protected String getParsedText() {
    return (matchNoneOfValues ? "not" : "") + Arrays.toString(values);
  }
}
